#ifndef _LINUX_SWAP_H
#define _LINUX_SWAP_H

#define SWAP_FLAG_PREFER	0x8000	/* set if swap priority specified */
#define SWAP_FLAG_PRIO_MASK	0x7fff
#define SWAP_FLAG_PRIO_SHIFT	0

#endif /* _LINUX_SWAP_H */
